/*
 * Copyright (c) 2001 Jason Tishler
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * $Id: rebase.cc,v 1.5 2001/12/18 20:21:19 jtishler Exp $
 */

#define DONT_USE_IMAGEHELP

#include <iostream>
#include <sstream>
#include <string>
#include <time.h>
#include <stdlib.h>
#include <getopt.h>
#include <sys/cygwin.h>
#include <windows.h>

#ifndef DONT_USE_IMAGEHELP
#include <imagehlp.h>
#endif 

#include <stdio.h>

using namespace std;

string PosixToWin32(const string& aPosixPath);
void ParseArgs(int argc, char* argv[]);
unsigned long StringToUlong(const string& aString);
void Usage();

ULONG theImageBase = 0;
BOOL theDownFlag = FALSE;
bool theDebugFlag = FALSE;
ULONG theOffset = 0;
int theArgsIndex = 0;
int theListFlag = 0; 

// lists Imagebase and Image size of a dll 

BOOL GetImageInfos(char *filename, uint *ImageBase, uint *ImageSize) 
{
	HANDLE hfile = CreateFile(filename, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL,0);
	if (hfile == INVALID_HANDLE_VALUE) {
		SetLastError(ERROR_FILE_NOT_FOUND); 
		return false; 
	}
	HANDLE hfilemapping = CreateFileMapping(hfile, NULL, PAGE_READONLY, 0, 0,  NULL);
	if (hfilemapping == 0) {
		CloseHandle(hfile);
		SetLastError(ERROR_FILE_NOT_FOUND); 
		return false; 
	}
	
	LPVOID lpFileBase = MapViewOfFile(hfilemapping, FILE_MAP_READ,0, 0, 0);
	if (lpFileBase == 0) {
		CloseHandle(hfilemapping);
		CloseHandle(hfile);
		SetLastError(ERROR_FILE_NOT_FOUND); 
		return false; 
	}

	PIMAGE_DOS_HEADER dosheader = (PIMAGE_DOS_HEADER)lpFileBase; 
	
	PIMAGE_NT_HEADERS ntheader = (PIMAGE_NT_HEADERS) ((char *)dosheader + dosheader->e_lfanew);

	*ImageBase = ntheader->OptionalHeader.ImageBase;
	*ImageSize = ntheader->OptionalHeader.SizeOfImage; 

	UnmapViewOfFile(lpFileBase);
	CloseHandle(hfilemapping);
	CloseHandle(hfile);
	SetLastError(0); 
	return true; 
}



BOOL ReBaseImage(
  PSTR CurrentImageName,  
  PSTR SymbolPath,      		// ignored 
  BOOL fReBase,          		
  BOOL fRebaseSysfileOk,  	// ignored 
  BOOL fGoingDown,       		// ignored 
  ULONG CheckImageSize,  		// ignored 
  ULONG *OldImageSize,  
  ULONG *OldImageBase,  
  ULONG *NewImageSize,  
  ULONG *NewImageBase,  
  ULONG TimeStamp        
)
{
	if (fReBase == 0) {
		SetLastError(ERROR_INVALID_PARAMETER);
		return false; 
	}
	
	HANDLE hfile = CreateFile(CurrentImageName, GENERIC_READ|GENERIC_WRITE, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL,0);
	if (hfile == INVALID_HANDLE_VALUE) {
		SetLastError(ERROR_FILE_NOT_FOUND); 
		return false; 
	}
	HANDLE hfilemapping = CreateFileMapping(hfile, NULL, PAGE_READWRITE, 0, 0,  NULL);
	if (hfilemapping == 0) {
		CloseHandle(hfile);
		SetLastError(ERROR_FILE_NOT_FOUND); 
		return false; 
	}
	
	LPVOID lpFileBase = MapViewOfFile(hfilemapping, FILE_MAP_WRITE,0, 0, 0);
	if (lpFileBase == 0) {
		CloseHandle(hfilemapping);
		CloseHandle(hfile);
		SetLastError(ERROR_FILE_NOT_FOUND); 
		return false; 
	}

	PIMAGE_DOS_HEADER dosheader = (PIMAGE_DOS_HEADER)lpFileBase; 
	
	PIMAGE_NT_HEADERS ntheader = (PIMAGE_NT_HEADERS) ((char *)dosheader + dosheader->e_lfanew);
	
	// set new header elements 
	*OldImageBase = ntheader->OptionalHeader.ImageBase; 
	*OldImageSize = ntheader->OptionalHeader.SizeOfImage; 
	*NewImageSize = ntheader->OptionalHeader.SizeOfImage; 

	// already rebased 
	if (ntheader->OptionalHeader.ImageBase == *NewImageBase) {
		SetLastError(0);
		return true; 
	}

	ntheader->OptionalHeader.ImageBase = *NewImageBase;
	ntheader->FileHeader.TimeDateStamp = TimeStamp; 

	if (theDebugFlag) printf("ImageBase: %08x ImageSize: %08x\n", ntheader->OptionalHeader.ImageBase,ntheader->OptionalHeader.SizeOfImage);

	PIMAGE_SECTION_HEADER	sections = (PIMAGE_SECTION_HEADER)	++ntheader;
	
	PIMAGE_BASE_RELOCATION relocs; 

	uint text_ofs;
	uint text_base;
	uint text_size;

	uint data_ofs;
	uint data_base;
	uint data_size;

	int diff = *NewImageBase - *OldImageBase; 
	if (theDebugFlag) printf("diff: %08x\n",diff);
	
	for (; *sections->Name != '\0'; ++sections) {
		if (theDebugFlag) {
			printf("section name %.8s\n",sections->Name); 
			printf("\toffset: 0x%8x\n",sections->PointerToRawData);
			printf("\tbase:   0x%8x\n",sections->VirtualAddress);
			printf("\tsize:   0x%8x\n",sections->SizeOfRawData);	
		}
		if (strncmp((char *)sections->Name,".reloc",6) == 0) {
				relocs = (PIMAGE_BASE_RELOCATION) ((unsigned int)sections->PointerToRawData + (unsigned int)lpFileBase); 
		}
		else if (strncmp((char *)sections->Name,".text",5) == 0) {
				text_ofs = sections->PointerToRawData;
				text_base = sections->VirtualAddress;
				text_size = sections->SizeOfRawData;
		}
		else if (strncmp((char *)sections->Name,".data",5) == 0) {
				data_ofs = sections->PointerToRawData;
				data_base = sections->VirtualAddress;
				data_size = sections->SizeOfRawData;
		}
	}	
	for (; relocs->SizeOfBlock != 0; ((char *)relocs) += relocs->SizeOfBlock  ) {
		int NumOfRelocs = (relocs->SizeOfBlock - sizeof(IMAGE_BASE_RELOCATION)) / sizeof (WORD); 
		int va = relocs->VirtualAddress; 
		PWORD p = (PWORD)((unsigned int )relocs + sizeof(IMAGE_BASE_RELOCATION));
		if (theDebugFlag) {
			printf("VirtAdress: %08x\n",relocs->VirtualAddress);
			printf("NumOfRelocs: %d\n",NumOfRelocs);
		}
		int offset; 

		if (va >= data_base)
			offset = data_ofs - data_base;
		else 
			offset = text_ofs - text_base;
		
		for (int i = 0; i < NumOfRelocs; i++,p++) {
			if ((*p & 0xf000) == 0x3000) {
				int location = (*p & 0x0fff) + va;
				if (theDebugFlag) {
					fprintf(stderr,"%08x - ",location);
					fprintf(stderr,"%08x\n",location + offset + 3);
				}
				int *patch_adr = (int *)(location + offset + (uint)lpFileBase); 
				*patch_adr +=	diff; 			
			}
		}
	}		

	UnmapViewOfFile(lpFileBase);
	CloseHandle(hfilemapping);
	CloseHandle(hfile);
	SetLastError(0);
	return true; 
}


void
main(int argc, char* argv[])
{
  ParseArgs(argc, argv);
  ULONG aNewImageBase = theImageBase;

  for (int i = theArgsIndex; i < argc; i++)
  {
    string aFile = PosixToWin32(argv[i]);
    if (theListFlag)
    {
    	uint ImageBase, ImageSize; 

			if (GetImageInfos(const_cast<char*>(aFile.c_str()),&ImageBase,&ImageSize) )
				printf("%-60s ImageBase: %08x ImageSize: %08x\n", const_cast<char*>(aFile.c_str()), ImageBase, ImageSize);
			else 
				printf("%-60s error: could not open file\n",const_cast<char*>(aFile.c_str()));
    }
    else {
    if (theDownFlag)
      aNewImageBase -= theOffset;

    ULONG anOldImageSize, anOldImageBase, aNewImageSize;
    ULONG aPrevNewImageBase = aNewImageBase;
    BOOL aStatus = ReBaseImage(
      const_cast<char*>(aFile.c_str()), // CurrentImageName
      0, // SymbolPath
      TRUE, // fReBase
      FALSE, // fRebaseSysfileOk
      theDownFlag, // fGoingDown
      0, // CheckImageSize
      &anOldImageSize, // OldImageSize
      &anOldImageBase, // OldImageBase
      &aNewImageSize, // NewImageSize
      &aNewImageBase, // NewImageBase
      time(0)); // TimeStamp

    // ReBaseImage seems to never returns false!
    DWORD aStatus2 = GetLastError();
    if (aStatus2 != 0)
    {
      cerr << "ReBaseImage(" << aFile.c_str() <<"," << hex << aNewImageBase <<") failed with last error = " <<
        dec << GetLastError() << endl;
      exit(2);
    }
    cout << aFile << hex << ": new base = " <<
      ((theDownFlag) ? aNewImageBase : aPrevNewImageBase) <<
      ", new size = " << aNewImageSize + theOffset << endl;

    if (!theDownFlag)
      aNewImageBase += theOffset;
    }
  }
  

  exit(0);
}

string
PosixToWin32(const string& aPosixPath)
{
  char aWin32Path[MAX_PATH];
  cygwin_conv_to_win32_path(aPosixPath.c_str(), aWin32Path);
  return aWin32Path;
}

void
ParseArgs(int argc, char* argv[])
{
  const char* anOptions = "b:Ddo:l:";
  for (int anOption; (anOption = getopt(argc, argv, anOptions)) != -1;)
  {
    switch (anOption)
    {      
    	case 'b':
        theImageBase = StringToUlong(optarg);
        break;
      case 'd':
        theDownFlag = TRUE;
        break;
      case 'o':
        theOffset = StringToUlong(optarg);
        break;
      case 'l':
        theListFlag = 1;
        break;
      case 'D':
        theDebugFlag= TRUE;
        break;
      default:
        Usage();
        exit(1);
        break;
    }
  }

  if (theImageBase == 0 && !theListFlag)
  {
    Usage();
    exit(1);
  }

  if (theListFlag)
    theArgsIndex = optind-1;
  else 
    theArgsIndex = optind;
}

unsigned long
StringToUlong(const string& aString)
{
  stringstream ss;
  unsigned long aUlong;
  string::size_type anIndex = aString.find("0x");
  if (anIndex == 0)
    ss << hex << string(aString, 2, aString.size() - 2);
  else
    ss << aString;
  ss >> aUlong;
  return aUlong;
}

void
Usage()
{
  cerr << "usage: rebase [-D] -b BaseAddress [-d] -o Offset <file> ...  rebase <file>  ([-D] print debug infos)" << endl;
  cerr << "usage: rebase -l <file> ...        list Imagebase and -size of <file>" << endl;
}

